## skr skr skr skr skr
def test_one(param1, param2=100):
    # loose comment
    pass

def test_one_half(pa: str) -> str:
    # test_one_half comment
    pass

def test_two_bevor():

    # above is intended to leave blank
    pass
# very close to previous comment
def test_two(arg: str):
    pass

def test_three(arg):
    """One line docstring."""
    pass

def test_four(param):
    '''Single quote docstring'''
    pass

def test_five(par):
    '''Two line single quote
    '''
    pass

def test_six():  # test_six right comment
    pass

# outside comment for test_seven
def test_seven():
    pass

class Foo():
    def __init__(self):
        pass
    
    def foo(self, word):
        # foo comment
        print(word)

    # bar comment
    def bar(self):
        pass

    def foobar(self): # foobar right comment
        pass



def perfect():
    """I'm perfect!"""
    print("I'm perfect!")

def test1():
    """
    """
    print("This is test1.")

def test2():
    # I'm test2.
    print("This is test2.")

# I'm test3.
def test3():
    print("This is test3.")

def test4(): # I'm test4.
    pass



# coding: utf-8
import os
import re
import tokenize as tk
from io import StringIO
from collections import deque
from typing import List


def get_content(filename: str) -> str:
    """Get raw content of a file."""
    with open(filename) as f:
        content = f.read()
        return content

def get_func_name(str_: str):
    prog = re.compile(r'\s*def\s+(.*)\(')
    result = prog.match(str_)
    return result.group(1) if result else result

def _check_cache(cache):
    """Check if cache contains def statement, if so, return name of the function, otherwise return None"""
    for line in cache:
        if 'def ' in line:
            return get_func_name(line)
    return None

def detect_infunc_and_same_line_comment(source: str) -> List:
    """Detect #-style comment in the first or second line inside a function.
    It can also detect #-style comment in the same line as def statement.

    e.g. functions like below can be detected:  
    def foo():
        # comment
        pass
    
    def bar():

        # comment
        pass
    
    def foobar():  # comment
        pass
    """
    func_name_and_cmt = []
    cache = deque(maxlen=2)
    for type_, token, _, _, line in tk.generate_tokens(StringIO(source).readline):
        if type_ != tk.COMMENT:
            cache.append(line)
        else:
            # check "def statement" in this line
            if 'def ' in line:
                func_name_and_cmt.append((get_func_name(line), token))
            else:
                # check "def statement" in preceding two lines (in cache)
                func_name = _check_cache(cache)
                if func_name:
                    func_name_and_cmt.append((func_name, token))
    return func_name_and_cmt

def detect_outfunc_comment(source: str) -> List:
    """Detect #-style comment outside a function.
    Note: the comment should be next to the function so that it can be detected.

    e.g. functions like below can be detected:
    # comment
    def foo():
        pass
    """
    func_name_and_cmt = []
    cache = deque(maxlen=1)
    for type_, token, _, _, line in tk.generate_tokens(StringIO(source).readline):
        if type_ not in [tk.NAME, tk.NL, tk.DEDENT]:
            cache.append((type_, token))
        else:
            if type_ == tk.NAME and token == 'def':
                prev_line_list = list(cache)
                if prev_line_list:
                    prev_line = prev_line_list[0]
                    if prev_line[0] == tk.COMMENT:
                        cmt = prev_line[1]
                        func_name = get_func_name(line)
                        func_name_and_cmt.append((func_name, cmt))
    return func_name_and_cmt

def get_all_names_of_funcs_with_comment(source: str) -> List:
    func_names = []
    infunc_and_same_cmt_list = detect_infunc_and_same_line_comment(source)
    outfunc_cmt_list = detect_outfunc_comment(source)
    cmt_list = infunc_and_same_cmt_list + outfunc_cmt_list
    for func_name, _ in cmt_list:
        func_names.append(func_name)
    return func_names

def get_py_files_inside_dir(path: str) -> List:
    """Return all python files' paths inside the given folder, recursively.
    Note:
    if you have subfolders inside the given folder, it also return python files' paths 
    inside the subfolders.
    """
    paths = []
    for root, _, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(".py"):
                paths.append(os.path.join(root, filename))
    return paths



from .main import get_ast_mod
from .main import get_ast_body
from .main import get_top_level_funcs
from .main import get_cls_nodes
from .main import get_cls_member_funcs
from .main import get_all_funcs
from .main import print_func_names
from .main import print_cls_names
from .main import check_func_docs
from .main import check_cls_docs
from .main import warn_func_docs_missing
from .main import warn_cls_docs_missing
from .utils import *

__version__ = '0.0.5'



#! /usr/bin/env python3
# coding: utf-8
import ast
import os
import itertools
import argparse
from typing import List, Generator
from .utils import *

def get_ast_mod(source: str) -> ast.Module:
    """Get ast module for the given python script.
    """
    mod = ast.parse(source)
    return mod

def get_ast_body(filename: str) -> List[ast.stmt]:
    """Get the body of ast module for the given python script.
    """
    mod = get_ast_mod(filename)
    return mod.body

def get_top_level_funcs(ast_body: List[ast.stmt]) -> Generator:
    top_level_funcs = (node for node in ast_body if isinstance(node, ast.FunctionDef))
    return top_level_funcs

def get_cls_nodes(ast_body: List[ast.stmt]) -> Generator:
    cls_nodes = (node for node in ast_body if isinstance(node, ast.ClassDef))
    return cls_nodes

def get_cls_member_funcs(ast_body: List[ast.stmt]) -> Generator:
    for cls_ in get_cls_nodes(ast_body):
        for cls_member_funcs in (node for node in cls_.body if isinstance(node, ast.FunctionDef)):
            yield cls_, cls_member_funcs

def get_all_funcs(ast_body: List[ast.stmt]) -> itertools.chain:
    top_level_funcs = get_top_level_funcs(ast_body)
    cls_member_funcs = get_cls_member_funcs(ast_body)
    all_funcs = itertools.chain(top_level_funcs, cls_member_funcs)
    return all_funcs

def print_func_names(func_nodes: Generator) -> None:
    """Here, we use if statements in for loop to differentiate two kind of func_nodes: class member functions
    and top-level functions. For class member functions, we organize them as tuple, (class_node, class_method_node),
    and for top-level functions, we organize them only as single values.
    """
    for func_node in func_nodes:
        if isinstance(func_node, tuple):  # if the node is a class member function node
            print("{}: {}".format(func_node[0].name, func_node[1].name))
        else:  # if the node is a top-level function node
            print(func_node.name)

def print_cls_names(cls_nodes: Generator) -> None:
    for cls_node in cls_nodes:
        print(cls_node.name)

def check_func_docs(func_nodes: Generator) -> None:
    """Here, func could either be a class member function or a top-level function, so we need to differentiate
    them using if statements in for loop.
    """
    for func_node in func_nodes:
        if isinstance(func_node, tuple):  # if the node is a class member function node
            # the output below is ("NAME_OF_CLASS: NAME_OF_CLASS_MEMBER_FUNCTION: DOC_OF_FUNCTION")
            print("{}: {}: {}".format(func_node[0].name, func_node[1].name, ast.get_docstring(func_node[1])))
        else:  # if the node is a top-level function node
            # the output below is ("NAME_OF_TOP_LEVEL_FUNCTION: DOC_OF_FUNCTION")
            print("{}: {}".format(func_node.name, ast.get_docstring(func_node)))

def check_cls_docs(cls_nodes: Generator) -> None:
    for cls_node in cls_nodes:
        # the output below is ("NAME_OF_CLASS: DOC_OF_CLASS")
        print("{}: {}".format(cls_node.name, ast.get_docstring(cls_node)))

def warn_func_docs_missing(func_nodes: Generator, name_list: List, strict: bool = False) -> None:
    for func_node in func_nodes:
        if isinstance(func_node, tuple):  # if the node is a class member function node
            if not ast.get_docstring(func_node[1]):
                if (not strict) and (func_node[1].name in name_list):
                    continue
                else:
                    print("WARNING: missing comments for functions in class: {}: {}".format(func_node[0].name, func_node[1].name))
        else:  # if the node is a top-level function node
            if not ast.get_docstring(func_node):
                if (not strict) and (func_node.name in name_list):
                    continue
                else:
                    print("WARNING: missing comments in top-level function: {}".format(func_node.name))

def warn_cls_docs_missing(cls_nodes: Generator) -> None:
    for cls_node in cls_nodes:
        if not ast.get_docstring(cls_node):
            print("WARNING: missing docstrings for class: {}".format(cls_node.name))

def _filename_check(filenames: List) -> bool:
    for filename in filenames:
        if not os.path.isfile(filename):
            return False
    return True

def init():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", nargs="+", help="path of folder or files")
    parser.add_argument("-s", "--strict", help="use strict mode", action="store_true")
    args = parser.parse_args()
    return (args.path, args.strict)

def main():
    path, strict_mode = init()
    if os.path.isdir(path[0]):  # if path is a list of folder names, we only handle the first folder
        filenames = get_py_files_inside_dir(path[0])
    else:
        if not _filename_check(path):
            print("Warning: Please indicate a valid path for file or folder.")
            return
        filenames = path

    mode = "strict mode" if strict_mode else "normal mode"

    for filename in filenames:
        print("({})====== {} ======:".format(mode, filename))
        source = get_content(filename)
        body = get_ast_body(source)
        func_nodes = get_all_funcs(body)
        cls_nodes = get_cls_nodes(body)
        warn_cls_docs_missing(cls_nodes)
        name_list = get_all_names_of_funcs_with_comment(source)
        warn_func_docs_missing(func_nodes, name_list, strict_mode)



"""Hello I'm testfile.py.
"""

def hello(who, when):
    print(who)
    print(when)

def world():
    """This is world.
    """
    print('World')

class HelloWorld():
    """This is HelloWorld class.
    """
    def __init__(self, name):
        self.name = name
    
    def test_no_args(self):
        print("This is test_no_args.")
    
    def test_with_args(self, what):
        print(what)
    
    def test_with_comment(self):
        """This is a comment of function: test_with_comment
        """
        pass

class JustDance(HelloWorld):
    def __init__(self, name):
        super().__init__(name)
    
    def just(self):
        print("This is just.")
    
    def dance(self, dc):
        print(dc)
    
    def test_with_comment(self, cm):
        """Just dance when you are not happy!
        """
        print(cm)
